require 'rubygems'
require 'open4'
require 'pp'
require 'json'
require 'openshift-origin-node'
require 'shellwords'
require 'facter'

module MCollective
  module Agent
    class Stickshift<RPC::Agent
      metadata    :name        => "Libra Management",
                  :description => "Agent to manage Libra services",
                  :author      => "Mike McGrath",
                  :license     => "ASL 2.0",
                  :version     => "0.1",
                  :url         => "http://www.openshift.com",
                  :timeout     => 240

      def echo_action
        validate :msg, String
        reply[:msg] = request[:msg]
      end

      def ss_app_create(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        app_uuid = args['--with-app-uuid']
        gear_uuid = args['--with-container-uuid']
        uid = args['--with-uid']
        uid = nil if uid && uid.to_s.empty?
        quota_blocks = args['--with-quota-blocks']
        quota_files = args['--with-quota-files']
        app_name = args['--with-app-name']
        gear_name = args['--with-container-name']
        namespace = args['--with-namespace']
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(app_uuid, gear_uuid, uid, app_name, gear_name,
                                                           namespace, quota_blocks, quota_files)
          container.create
        rescue StickShift::UserCreationException => e
          Log.instance.info e.message
          return 129, e.message
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_app_destroy(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"
        app_uuid = args['--with-app-uuid']
        uuid = args['--with-container-uuid']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(app_uuid, uuid)
          container.destroy
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_authorized_ssh_key_add(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        uuid = args['--with-container-uuid']
        app_uuid = args['--with-app-uuid']
        ssh_key = args['--with-ssh-key']
        key_type = args['--with-ssh-key-type']
        comment = args['--with-ssh-key-comment']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(uuid, uuid)
          container.user.add_ssh_key(ssh_key, key_type, comment)
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_authorized_ssh_key_remove(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        uuid = args['--with-container-uuid']
        app_uuid = args['--with-app-uuid']
        ssh_key = args['--with-ssh-key']
        comment = args['--with-ssh-comment']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(uuid, uuid)
          container.user.remove_ssh_key(ssh_key, comment)
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_broker_auth_key_add(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        uuid = args['--with-container-uuid']
        app_uuid = args['--with-app-uuid']
        iv = args['--with-iv']
        token = args['--with-token']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(uuid, uuid)
          container.user.add_broker_auth(iv, token)
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_broker_auth_key_remove(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        uuid = args['--with-container-uuid']
        app_uuid = args['--with-app-uuid']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(uuid, uuid)
          container.user.remove_broker_auth
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_env_var_add(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        uuid = args['--with-container-uuid']
        app_uuid = args['--with-app-uuid']
        key = args['--with-key']
        value = args['--with-value']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(uuid, uuid)
          container.user.add_env_var(key, value)
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_env_var_remove(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        uuid = args['--with-container-uuid']
        key = args['--with-key']
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(uuid, uuid)
          container.user.remove_env_var(key)
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_cartridge_list(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        list_descriptors = true if args['--with-descriptors']
        porcelain = true if args['--porcelain']

        output = ""
        begin
          output = StickShift::Node.get_cartridge_list(list_descriptors, porcelain, false)
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_app_state_show(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"

        container_uuid = args['--with-container-uuid']
        app_uuid = args['--with-app-uuid']        
        
        output = ""
        begin
          container = StickShift::ApplicationContainer.new(app_uuid, container_uuid)
          output = container.get_app_state()
        rescue Exception => e
          Log.instance.info e.message
          return -1, e.message
        else
          return 0, output
        end
      end

      def ss_connector_execute(cmd, args)
        Log.instance.info "COMMAND: #{cmd}"
        gear_uuid = args['--gear-uuid']
        cart_name = args['--cart-name']
        hook_name = args['--hook-name']
        input_args = args['--input-args']
        
        hook_path = "/usr/libexec/stickshift/cartridges/#{cart_name}/info/connection-hooks/#{hook_name}"
        if File.exists? hook_path
           pid, stdin, stdout, stderr = Open4::popen4ext(true, "#{hook_path} #{input_args} 2>&1")
        else
           raise Exception.new("Could not find #{hook_path}")
        end
        return pid, stdin, stdout, stderr
      end

      def handle_ss_cmd(action, args)
        cmd = "ss-#{action}"
        case action
        when "app-create"
          rc, output = ss_app_create(cmd, args)
        when "app-destroy"
          rc, output = ss_app_destroy(cmd, args)
        when "authorized-ssh-key-add"
          rc, output = ss_authorized_ssh_key_add(cmd, args)
        when "authorized-ssh-key-remove"
          rc, output = ss_authorized_ssh_key_remove(cmd, args)
        when "broker-auth-key-add"
          rc, output = ss_broker_auth_key_add(cmd, args)
        when "broker-auth-key-remove"
          rc, output = ss_broker_auth_key_remove(cmd, args)
        when "env-var-add" 
          rc, output = ss_env_var_add(cmd, args)
        when "env-var-remove" 
          rc, output = ss_env_var_remove(cmd, args)
        when "cartridge-list"
          rc, output = ss_cartridge_list(cmd, args)
        when "app-state-show"
          rc, output = ss_app_state_show(cmd, args)
        else
          return nil, nil
        end
        return rc, output
      end

      #
      # Passes arguments to cartridge for use
      #
      def cartridge_do_action
        Log.instance.info("cartridge_do_action call / request = #{request.pretty_inspect}")
        Log.instance.info("cartridge_do_action validation = #{request[:cartridge]} #{request[:action]} #{request[:args]}")
        validate :cartridge, /\A[a-zA-Z0-9\.\-\/]+\z/
        validate :cartridge, :shellsafe
        validate :action, /\A(app-create|app-destroy|env-var-add|env-var-remove|broker-auth-key-add|broker-auth-key-remove|authorized-ssh-key-add|authorized-ssh-key-remove|configure|deconfigure|update-namespace|tidy|deploy-httpd-proxy|remove-httpd-proxy|move|pre-move|post-move|info|post-install|post-remove|pre-install|reload|restart|start|status|stop|force-stop|add-alias|remove-alias|threaddump|cartridge-list|expose-port|conceal-port|show-port|system-messages|connector-execute)\Z/
        validate :action, :shellsafe
        cartridge = request[:cartridge]
        action = request[:action]
        args = request[:args]
        pid, stdin, stdout, stderr = nil, nil, nil, nil
        rc = nil
        output = ""
        if cartridge == 'stickshift-node'
          cmd = "ss-#{action}"
          if action == 'connector-execute'
            pid, stdin, stdout, stderr = ss_connector_execute(cmd, args)
          else
            rc, output = handle_ss_cmd(action, args)
            reply[:output] = output
            reply[:exitcode] = rc
            if rc == 0
              Log.instance.info("cartridge_do_action (#{rc})\n------\n#{output}\n------)")
            else
              Log.instance.info("cartridge_do_action ERROR (#{rc})\n------\n#{output}\n------)")
            end
            reply.fail! "cartridge_do_action failed #{rc}.  Output #{output}" unless rc == 0
            return 0
          end
        else
          validate :args, /\A[\w\+\/= \{\}\"@\-\.:;\'\\\n~,]+\z/
          validate :args, :shellsafe
          if File.exists? "/usr/libexec/stickshift/cartridges/#{cartridge}/info/hooks/#{action}"                
            pid, stdin, stdout, stderr = Open4::popen4ext(true, "/usr/bin/runcon -l s0-s0:c0.c1023 /usr/libexec/stickshift/cartridges/#{cartridge}/info/hooks/#{action} #{args} 2>&1")
          elsif File.exists? "/usr/libexec/stickshift/cartridges/embedded/#{cartridge}/info/hooks/#{action}"                
            pid, stdin, stdout, stderr = Open4::popen4ext(true, "/usr/bin/runcon -l s0-s0:c0.c1023 /usr/libexec/stickshift/cartridges/embedded/#{cartridge}/info/hooks/#{action} #{args} 2>&1")
          else
            reply[:exitcode] = 127
            reply.fail! "cartridge_do_action ERROR action '#{action}' not found."
          end
        end
        stdin.close
        ignored, status = Process::waitpid2 pid
        exitcode = status.exitstatus
        # Do this to avoid cartridges that might hold open stdout
        output = ""
        begin
          Timeout::timeout(5) do
            while (line = stdout.gets)
              output << line
            end
          end
        rescue Timeout::Error
          Log.instance.info("cartridge_do_action WARNING - stdout read timed out")
        end

        if exitcode == 0
          Log.instance.info("cartridge_do_action (#{exitcode})\n------\n#{output}\n------)")
        else
          Log.instance.info("cartridge_do_action ERROR (#{exitcode})\n------\n#{output}\n------)")
        end

        reply[:output] = output
        reply[:exitcode] = exitcode
        reply.fail! "cartridge_do_action failed #{exitcode}.  Output #{output}" unless exitcode == 0
      end
     
      #
      # Set the district for a node
      #
      def set_district_action
        Log.instance.info("set_district call / request = #{request.pretty_inspect}")
        validate :uuid, /^[a-zA-Z0-9]+$/
        uuid = request[:uuid]
        active = request[:active]

        begin
          district_home = '/var/lib/stickshift/.settings'
          FileUtils.mkdir_p(district_home)

          File.open(File.join(district_home, 'district.info'), 'w') { |f|
            f.write("#Do not modify manually!\nuuid='#{uuid}'\nactive='#{active}'\n")
          }

          Facter.add(:district_uuid) do
            setcode { uuid }
          end
          Facter.add(:district_active) do
            setcode { active }
          end

          reply[:output] = "created/updated district #{uuid} with active = #{active}"
          reply[:exitcode] = 0
        rescue Exception => e
          reply[:output] = e.message
          reply[:exitcode] = 255
          reply.fail! "set_district failed #{reply[:exitcode]}.  Output #{reply[:output]}" 
        end

        Log.instance.info("set_district (#{reply[:exitcode]})\n------\n#{reply[:output]}\n------)")
      end

      #
      # Returns whether an app is on a server
      #
      def has_app_action
        validate :uuid, /^[a-zA-Z0-9]+$/
        validate :application, /^[a-zA-Z0-9]+$/
        uuid = request[:uuid]
        app_name = request[:application]
        if File.exist?("/var/lib/stickshift/#{uuid}/#{app_name}")
          reply[:output] = true
        else
          reply[:output] = false
        end
        reply[:exitcode] = 0
      end
      
      #
      # Returns whether an embedded app is on a server
      #
      def has_embedded_app_action
        validate :uuid, /^[a-zA-Z0-9]+$/
        validate :embedded_type, /^.+$/
        uuid = request[:uuid]
        embedded_type = request[:embedded_type]
        if File.exist?("/var/lib/stickshift/#{uuid}/#{embedded_type}")
          reply[:output] = true
        else
          reply[:output] = false
        end
        reply[:exitcode] = 0
      end
      
      #
      # Returns whether a uid or gid is already reserved on the system
      #
      def has_uid_or_gid_action
        validate :uid, /^[0-9]+$/
        uid = request[:uid].to_i

        uids = IO.readlines("/etc/passwd").map{ |line| line.split(":")[2].to_i }
        gids = IO.readlines("/etc/group").map{ |line| line.split(":")[2].to_i }

        if uids.include?(uid) || gids.include?(uid)
          reply[:output] = true
        else
          reply[:output] = false
        end
        reply[:exitcode] = 0
      end

      #
      # Get all gears
      #
      def get_all_gears_action
        gear_map = request[:gear_map]

        uid_map = {}
        uids = IO.readlines("/etc/passwd").map{ |line| 
          uid = line.split(":")[2]
          username = line.split(":")[0]
          uid_map[username] = uid
        }
        dir = "/var/lib/stickshift/"
        filelist = Dir.foreach(dir) { |file| 
          if File.directory?(dir+file) and not File.symlink?(dir+file) and not file[0]=='.'
            if uid_map.has_key?(file)
              gear_map[file] = uid_map[file]
            end
          end
        }
        reply[:output] = gear_map
        reply[:exitcode] = 0
      end


      def handle_ss_job(parallel_job)
        job = parallel_job[:job]
        rc, output = handle_ss_cmd(job[:action], job[:args])
        parallel_job[:result_exit_code] = rc
        if rc == 0
          parallel_job[:result_stdout] = output
          parallel_job[:result_stderr] = ""
        else
          parallel_job[:result_stdout] = ""
          parallel_job[:result_stderr] = output
        end
      end

      #
      # Executes a list of jobs parallely and returns their results embedded in args
      #
      def execute_parallel_action        
        Log.instance.info("execute_parallel_action call / request = #{request.pretty_inspect}")
        #validate :joblist, /\A[\w\+\/= \{\}\"@\-\.:\'\\\n~,_]+\z/
        #validate :joblist, :shellsafe

        joblist = request[config.identity]
        pidlist = []
        inline_list = []
        joblist.each { |parallel_job|
          job = parallel_job[:job]
          cartridge = job[:cartridge]
          action = job[:action]
          args = job[:args]
          if cartridge == 'stickshift-node' && action != 'connector-execute'
            inline_list << parallel_job
          else
            begin
              if cartridge == 'stickshift-node' && action == 'connector-execute'
                pid, stdin, stdout, stderr = ss_connector_execute(action, args)
              else
                pid, stdout, stderr = execute_parallel_job(cartridge, action, args)
              end
            rescue Exception =>e
              parallel_job[:result_exit_code] = 127
              parallel_job[:result_stdout] = e.message
              parallel_job[:result_stderr] = e.message
              next
            end
            pidlist << [parallel_job, pid, stdout, stderr]
          end
        }

        # All the inline calls are made using multiple threads instead of processes
        in_threads = []
        inline_list.each do |parallel_job|
          in_threads << Thread.new(parallel_job) do |pj|
            begin
              handle_ss_job(pj)
            rescue Exception => e
              pj[:result_exit_code] = 1
              pj[:result_stdout] = e.message
              pj[:result_stderr] = e.message
              next
            end
          end
        end
        in_threads.each { |thr| thr.join }

        pidlist.each { |reap_args|
          pj, pid, sout, serr = reap_args
          reap_output(pj, pid, sout, serr)
        }
        Log.instance.info("execute_parallel_action call - 10 #{joblist}")
        reply[:output] = joblist
        reply[:exitcode] = 0
      end

      def execute_parallel_job(cartridge, action, args)
        pid, stdin, stdout, stderr = nil, nil, nil, nil
        if cartridge == 'stickshift-node' && action == 'connector-execute'
          cmd = "ss-#{action}"
          pid, stdin, stdout, stderr = Open4::popen4("/usr/bin/runcon -l s0-s0:c0.c1023 #{cmd} #{args} 2>&1")
        else
          if File.exists? "/usr/libexec/stickshift/cartridges/#{cartridge}/info/hooks/#{action}"                
            pid, stdin, stdout, stderr = Open4::popen4ext(true, "/usr/bin/runcon -l s0-s0:c0.c1023 /usr/libexec/stickshift/cartridges/#{cartridge}/info/hooks/#{action} #{args} 2>&1")
            #pid, stdin, stdout, stderr = Open4::popen4("/usr/bin/runcon -l s0-s0:c0.c1023 /usr/libexec/stickshift/cartridges/#{cartridge}/info/hooks/#{action} #{args} 2>&1")
          elsif File.exists? "/usr/libexec/stickshift/cartridges/embedded/#{cartridge}/info/hooks/#{action}"                
            pid, stdin, stdout, stderr = Open4::popen4ext(true, "/usr/bin/runcon -l s0-s0:c0.c1023 /usr/libexec/stickshift/cartridges/embedded/#{cartridge}/info/hooks/#{action} #{args} 2>&1")
          else
            raise Exception.new("cartridge_do_action ERROR action '#{action}' not found.")
          end
        end
        stdin.close
        return pid, stdout, stderr
      end

      def reap_output(parallel_job, pid, stdout, stderr)
        ignored, status = Process::waitpid2 pid
        exitcode = status.exitstatus
        # Do this to avoid cartridges that might hold open stdout
        output = ""
        begin
          Timeout::timeout(5) do
            while (line = stdout.gets)
              output << line
            end
          end
        rescue Timeout::Error
          Log.instance.info("cartridge_do_action WARNING - stdout read timed out")
        end

        if exitcode == 0
          Log.instance.info("cartridge_do_action (#{exitcode})\n------\n#{output}\n------)")
        else
          Log.instance.info("cartridge_do_action ERROR (#{exitcode})\n------\n#{output}\n------)")
        end

        parallel_job[:result_stdout] = output
        parallel_job[:result_exit_code] = exitcode
      end
    end
  end
end
